import sys
import os
import re
import logging
import time
import json
import subprocess
import traceback
import tempfile
from pathlib import Path

from ktpanda.textcolor import Colorizer

from .constants import DEFAULT_CONFIG
from .database import BackupDB
from .utils import (command, arg, canonical_path, escape_like, prefix_upper,
                    addcomma, prettysize, round_blocks, parsesize, restore_mtime)
from .nexus import NexusLevelDict

log = logging.getLogger(__name__)

##########################################################################################
# Completers
##########################################################################################

def complete_virtual_path(prefix, parsed_args, **kwargs):
    db = open_db(parsed_args, expand=True)
    prefix = canonical_path(prefix, None)
    pflen = len(prefix)
    rtn_paths = {}
    rx_stripslash = re.compile(r'/.*')
    for row in db.execute("SELECT virtual_path FROM file_tree WHERE virtual_path >= ? AND virtual_path < ?", (prefix, prefix_upper(prefix))):
        path = row[0]
        path = path[:pflen] + rx_stripslash.sub('/', path[pflen:])
        rtn_paths[path] = True

    return list(rtn_paths)

def complete_disk(prefix, parsed_args, **kwargs):
    db = open_db(parsed_args, expand=True)
    if '/' in prefix:
        return []

    rtn = []
    like_prefix = escape_like(prefix) + '%'
    for row in db.execute("SELECT name FROM disk WHERE name NOTNULL AND name LIKE ? ESCAPE '\\'", (like_prefix,)):
        rtn.append(row[0])

    return rtn

##########################################################################################
# Utilities
##########################################################################################

def open_db(args, expand=False):
    if args.db and expand:
        args.db = args.db.expanduser()
    db = BackupDB(args.db, args.noaction, args.verbose)
    db.nocopydb = args.nocopy
    db.explain = args.explain
    db.connect()
    return db

##########################################################################################
# Common arguments
##########################################################################################

arg_disk = arg('disk', metavar='NAME', completer=complete_disk, help='Name, UUID, or path to disk data files')
arg_update = arg('-u', '--update', action='store_true', help='Perform an update before copying')
arg_nocopy = arg('--nocopy', action='store_true', help='Do not copy database')
arg_nohelper = arg('--nohelper', action='store_true', help='Do not use hash_helper module')
arg_sync = arg('--sync', action='store_true', help='Sync files to disk after copy')
arg_ignore_lock = arg('--ignore-lock', action='store_true', help='Scan locked directories')

@command('info', help='Report usage')
@arg('--no-saturated', action='store_true', help='Exclude the size of saturated objects from individual disk counts')
def cmd_report(db:BackupDB, parser, args):
    from .report import Reporter
    Reporter(db).print_report(exclude_saturated=args.no_saturated)

@command('ls', help='List files in database')
@arg('filter', metavar='PATTERN', nargs='?', completer=complete_virtual_path, help='Limit view to files that match the given pattern')
@arg('-d', '--disk', metavar='NAME', completer=complete_disk, help='Limit view to files on the given disk')
def cmd_list(db:BackupDB, parser, args):
    from .report import Reporter
    if args.disk:
        db.set_backup_disk(args.disk)
    Reporter(db).list_files(args.filter)

@command('tree', help='Show tree of files')
def cmd_tree(db:BackupDB, parser, args):
    from .report import Reporter
    root = Reporter(db).get_tree()
    def dumpnode(level, node):
        print('%02d %5d %18s %s %2s %s %s' % (level, node.total_files, addcomma(node.total_size), ' '.join('%4d/%-8s' % (fcnt, prettysize(size)) for fcnt, size in zip(node.copy_files, node.copy_size)), '' if node.copies is None else node.copies, '   ' * level, node.name))
        for cnode in sorted(list(node.children.values()), key=lambda c:-c.total_size):
            dumpnode(level + 1, cnode)
    dumpnode(0, root)

@command('ncdu', help='Show size of backup set using `ncdu`')
@arg('-o', '--output', type=Path, metavar='FILE', help='Write output to file (readable by `ncdu -f`)')
def cmd_ncdu(db:BackupDB, parser, args):
    from .report import Reporter
    data = Reporter(db).make_ncdu()
    if args.output:
        with args.output.open('w') as fp:
            json.dump(data, fp, indent=True)
    else:
        subprocess.run(['ncdu', '-f', '-'], encoding='utf8', input=json.dumps(data))


##########################################################################################
# Update
##########################################################################################

@command('update', help='Update the database view of the source files')
@arg_ignore_lock
@arg_nohelper
def cmd_update(db:BackupDB, parser, args):
    from .update import Updater
    from .report import Reporter
    Updater(db).update_source(args.ignore_lock)
    log.info('source updated.')
    Reporter(db).print_report()

##########################################################################################
# Backup
##########################################################################################

@command('backup', help='Back up files to a disk')
@arg_disk
@arg('--slack', action='store', help='Extra temporary space on target')
@arg('--limit', action='store', help='Stop after copying a specific amount of data')
@arg('--maxcopies', action='store', metavar='N', type=int, help="Don't make new copies of objects that already have at least N copies")
@arg('--noflush', action='store_true', help='Do not delete redundant copies to make room for new files')
@arg('--simulate', action='store_true', help='Do not actually modify target')
@arg_sync
@arg_update
@arg_ignore_lock
@arg_nohelper
def cmd_backup(db:BackupDB, parser, args):
    from .backup import Backup
    from .report import Reporter
    db.set_backup_disk(args.disk)
    if args.update:
        from .update import Updater
        Updater(db).update_source(args.ignore_lock)
        log.info('source updated.')

    limit = None
    if args.limit is not None:
        limit = parsesize(args.limit)

    pre_levels = NexusLevelDict(db.load_nexus())
    levels = Backup(db).backup_files(
        simulate=args.simulate,
        noflush=args.noflush,
        limit_bytes=limit,
        limit_copies=args.maxcopies
    )
    log.info('Pre-backup state:')
    reporter = Reporter(db)
    reporter.print_report(pre_levels, indent='    ')
    log.info('')
    log.info('Post-backup state:')
    reporter.print_report(levels, indent='    ')
    if args.simulate:
        log.info(f'simulated update to {db.current_disk.name} complete.')
    else:
        log.info(f'destination {db.current_disk.name} updated.')


##########################################################################################
# Restore
##########################################################################################

@command('restore', help='Restore files from backup')
@arg('src', nargs='?', metavar='DISK', completer=complete_disk, help='Disk or path to restore from')
@arg('dest', nargs='?', type=Path, metavar='PATH', help='Destination path to copy restored files')
@arg('-l', '--link', action='store_true', help='Make a hard link of the files instead of copying')
@arg('-f', '--filter', metavar='PATTERN', completer=complete_virtual_path, help='Restore only files that match PATTERN')
@arg('-e', '--existing', action='append', default=[], help='Ignore files that are present relative to this folder (can be specified multiple times)')
@arg('--minset', action='store_true', help="Deterimine the minimum set of disks necessary to restore the files")
def cmd_restore(db:BackupDB, parser, args):
    from .restore import Restore
    restore = Restore(db, args.dest)
    if args.minset:
        restore.restore_set(args.filter, args.existing)
        return

    if args.src is None or args.dest is None:
        parser.print_help()
        return

    db.set_backup_disk(args.src)

    if args.link:
        restore.restore_link(args.filter, args.existing)
    else:
        restore.restore_copy(args.filter, args.existing)

##########################################################################################
# Verify
##########################################################################################

@command('verify', help='Verify the integrity of the files on a backup disk')
@arg_disk
def cmd_verify(db:BackupDB, parser, args):
    db.set_backup_disk(args.disk)
    db.verify()

##########################################################################################
# Source
##########################################################################################

@command('source', help='Manage backup sources')
def list_sources(db:BackupDB, parser, args):
    if not db.path_map:
        print('No sources defined')
        return

    vheader = 'Virtual path'
    pheader = 'Physical path'
    vpath_len = max(len(vheader), max(len(vpath) for vpath in db.path_map))
    ppath_len = max(len(pheader), max(len(str(ppath)) for ppath in db.path_map.values()))
    print(f'{vheader.ljust(vpath_len)} {pheader}')
    print(f'{"-" * vpath_len} {"-" * ppath_len}')
    for vpath, ppath in db.path_map.items():
        print(f'{vpath.ljust(vpath_len)} {ppath}')


@command('source', 'map', help='Map a virtual path to a physical path')
@arg('vpath', metavar='VPATH', completer=complete_virtual_path, help='Virtual path in backup set')
@arg('ppath', metavar='PATH', type=Path, help='Physical path where files are located')
def source_map(db:BackupDB, parser, args):
    vpath = canonical_path(args.vpath, True)
    ppath = args.ppath
    if not ppath.is_absolute():
        ppath = ppath.resolve()

    db.map_source(vpath, str(ppath))
    print(f'Set {vpath} -> {ppath}')

@command('source', 'unmap', aliases=['del'], help='Delete a virtual path mapping')
@arg('vpath', metavar='VPATH', completer=complete_virtual_path, help='Virtual path in backup set')
def del_source(db:BackupDB, parser, args):
    vpath = canonical_path(args.vpath, True)
    if db.unmap_source(vpath):
        print(f'Deleted path {vpath}')
    else:
        print(f'Virtual path {vpath} not found')

@command('source', 'lock', help='Lock source paths (prevent scanning, assume unchanged)')
@arg('vpath', nargs='*', metavar='VPATH', completer=complete_virtual_path, help='Virtual path in backup set')
def lock_source(db:BackupDB, parser, args, lock=True):
    if not args.vpath:
        for vpath, config in db.execute('SELECT virtual_path, config FROM path_config'):
            if config and json.loads(config).get('lock'):
                print(vpath)

    for vpath, prev_config, new_config in db.set_path_config(args.vpath, {'lock': lock}):
        vpath = canonical_path(vpath, True)
        was_locked = bool(prev_config.get('lock'))
        if was_locked:
            print(f'{vpath} already locked' if lock else f'Unlocked: {vpath}')
        else:
            print(f'Locked: {vpath}' if lock else f'{vpath} not locked')

@command('source', 'unlock', help='Unlock source paths')
@arg('vpath', nargs='+', metavar='VPATH', completer=complete_virtual_path, help='Virtual path in backup set')
def unlock_source(db:BackupDB, parser, args):
    return lock_source(db, parser, args, False)

@command('source', 'set', help='Configure source paths')
@arg('vpath', nargs='+', metavar='VPATH', completer=complete_virtual_path, help='Virtual path in backup set')
@arg('-j', '--json', metavar='JSON', help='Specify updates in JSON format. Other options will override fields from this.')
@arg('-l', '--lock', action='store_true', help='Lock the specified paths')
@arg('-u', '--unlock', action='store_true', help='Unlock the specified paths')
@arg('-i', '--ignore', metavar='PATTERNS', action='append', default=None,
     help='Filename or patterns to ignore. Prefix with `!` to negate the ignore '
     '(include even if excluded by a later rule). The first pattern that matches '
     'determines the action, so include rules should be specified before ignore rules')
@arg('-a', '--append', action='store_true', help='Append new rules instead of replacing existing rules')
@arg('-p', '--priority', type=int, help='Set the default priority for the given paths (use 0 to clear)')
@arg('-m', '--maxcopies', type=int, help='Set the default priority for the given paths (use 0 to clear)')
@arg('--hide', action='store_true', help='Mark the specified paths as hidden')
@arg('--unhide', action='store_true', help='Unhide the specified paths')
@arg('--follow', action='store_true', help='Follow symbolic links in the directory')
@arg('--nofollow', action='store_true', help='Do not follow symbolic links in the directory')
def set_source(db:BackupDB, parser, args):
    updates = json.loads(args.json) if args.json else {}
    if args.lock:
        updates['lock'] = True
    if args.unlock:
        updates['lock'] = False
    if args.hide:
        updates['hide'] = True
    if args.unhide:
        updates['hide'] = None
    if args.follow:
        updates['follow_symlinks'] = True
    if args.nofollow:
        updates['follow_symlinks'] = None

    if args.priority is not None:
        updates['priority'] = args.priority or None

    if args.maxcopies is not None:
        updates['maxcopies'] = args.maxcopies or None

    lst = args.ignore
    if lst is not None:
        update_list = []
        for val in lst:
            update_list.append(val.strip())
        updates['rules'] = update_list

    if not updates:
        for vpath in args.vpath:
            vpath = canonical_path(vpath, True)
            config_text = db.query_scalar('SELECT config FROM path_config WHERE virtual_path = ?', (vpath,))
            config = dict(DEFAULT_CONFIG)
            if config_text:
                config.update(json.loads(config_text))
            print(f'{vpath}: {json.dumps(config)}')
        return

    print(f'Applying the following configuration changes: {json.dumps(updates)}')
    for vpath, prev_config, new_config in db.set_path_config(args.vpath, updates, False, ('rules',) if args.append else ()):
        all_keys = prev_config.keys() | new_config.keys()
        changes = {}
        for key in sorted(all_keys):
            default = DEFAULT_CONFIG.get(key)
            prev_val = prev_config.get(key, default)
            new_val = new_config.get(key, default)
            if prev_val != new_val:
                changes[key] = prev_val, new_val
        if changes:
            print(f'{vpath}:')
            for key, (prev_val, new_val) in changes.items():
                print(f'   {key}: {prev_val} -> {new_val}')
            print()
        else:
            print(f'{vpath}: no changes')

@command('source', 'reset', help='Clear configuration for source paths')
@arg('vpath', nargs='+', metavar='VPATH', completer=complete_virtual_path, help='Virtual path in backup set')
def reset_source(db:BackupDB, parser, args):
    paths = [canonical_path(path, True) for path in args.vpath]
    db.set_path_config(paths, {}, True)

##########################################################################################
# Disk
##########################################################################################

@command('disk', help='Manage backup disks')
def cmd_disk(db:BackupDB, parser, args):
    parser.print_help()

@command('disk', 'list', help='List all disks')
def cmd_disk_list(db:BackupDB, parser, args):
    db.find_disk_data_paths()
    clridx = 1
    print(f'{"disk":<20} {"device":<20} data path')
    maxpath = max(len(str(disk.data_path or '')) for disk in db.disk_by_name.values())
    for name, disk in sorted(db.disk_by_name.items()):
        txt = Colorizer(fg='ccc' if (clridx & 1) else 'fff', bg='222' if (clridx & 1) else '333')
        txt.t(name, rpad=20).t(' ')
        txt.t(disk.device_name or "-", rpad=20).t(' ')
        txt.t(str(disk.data_path or "-"), rpad=maxpath)
        print(txt.get())
        clridx += 1

@command('disk', 'add', help='Add a backup disk')
@arg('name', metavar='NAME', help='Name of new disk')
@arg('path', metavar='PATH', type=Path, help='Path to data files')
@arg('-s', '--size', help='Set size of selected disk')
@arg('-p', '--path', metavar='PATH', help='Path to data files')
def cmd_disk_add(db:BackupDB, parser, args):
    args.path.mkdir(parents=True, exist_ok=True)
    path = args.path.resolve()
    disk = db.add_disk(args.name, path, parsesize(args.size))
    db.copy_database()
    print(f'Added disk {disk.name}:')
    print(f'  UUID: {disk.uuid}')
    print(f'  Nexus index: {disk.nexus_index}')
    print(f'  Size: {addcomma(disk.size)}')
    if disk.device_name:
        print(f'  Device (current): {disk.device_name}')
    if disk.mountpoint:
        print(f'  Mount point (current): {disk.mountpoint}')
    if disk.relative_path:
        print(f'  Relative path from mountpoint: {disk.relative_path}')

@command('disk', 'set', help='Set parameters for a backup disk')
@arg_disk
@arg('-s', '--size', help='Set size of selected disk')
@arg('-p', '--path', metavar='PATH', type=Path, help='Path to data files')
@arg('-r', '--relpath', help='Set path of data files relative to mount')
def cmd_disk_set(db:BackupDB, parser, args):
    db.set_backup_disk(args.disk)
    db.set_disk_params(size=parsesize(args.size), relpath=args.relpath, path=args.path)

@command('disk', 'rename', aliases=['mv'], help='Set parameters for a backup disk')
@arg_disk
@arg('newname', metavar='NAME', help='New name for disk')
def cmd_disk_rename(db:BackupDB, parser, args):
    db.set_backup_disk(args.disk)
    db.rename_disk(args.newname)

@command('disk', 'drop', help='Drop all files from a disk. Use when a disk has failed.')
@arg_disk
def disk_drop(db:BackupDB, parser, args):
    db.set_backup_disk(args.disk)
    db.drop_disk()

@command('disk', 'del', help='Delete a disk. The disk must not have any files (use `disk drop` first)')
@arg_disk
def disk_del(db:BackupDB, parser, args):
    db.set_backup_disk(args.disk)
    db.delete_disk()

@command('disk', 'refresh', help='Refresh database view of files on the selected disk.')
@arg_disk
def disk_refresh(db:BackupDB, parser, args):
    db.set_backup_disk(args.disk)
    db.refresh_disk()
    log.info('database refreshed.')

##########################################################################################
# Config
##########################################################################################

@command('config', help='Manage configuration')
def cmd_config(db:BackupDB, parser, args):
    parser.print_help()

@command('config', 'export', help='Export configuration to JSON file')
@arg('path', type=Path, help='File to export to, or "-" for standard output')
@arg('-p', '--path-prefix', completer=complete_virtual_path, help='Export path_config only for the given directory and subdirectories')
def export_config(db:BackupDB, parser, args):
    data = db.export_config(canonical_path(args.path_prefix, True))
    if str(args.path) == '-':
        print(json.dumps(data, indent=4))
    else:
        with args.path.open('w') as fp:
            json.dump(data, fp, indent=4)

@command('config', 'import', help='Import configuration from JSON file')
@arg('path', type=Path, help='File to import from, or "-" for standard output')
def import_config(db:BackupDB, parser, args):
    if str(args.path) == '-':
        data = json.load(sys.stdin)
    else:
        with args.path.open() as fp:
            data = json.load(fp)
    db.import_config(data)

@command('config', 'edit', help='Edit configuration with an external editor')
@arg('-e', '--editor', help='Editor command (defaults to $EDITOR environment variable)')
def edit_config(db:BackupDB, parser, args, keys=(), path_prefix=None):
    editor = os.getenv('EDITOR') or 'editor'
    if args.editor:
        editor = args.editor

    with tempfile.NamedTemporaryFile(prefix='distbackup-edit-', suffix='.json') as tmpf:
        path = Path(tmpf.name)
        data = db.export_config(path_prefix)
        if keys:
            data = {k: v for k, v in data.items() if k in keys}

        with path.open('w', encoding='utf8') as fp:
            json.dump(data, fp, indent=4)

        while True:
            print(f'Running editor: {editor}')
            try:
                subprocess.run([editor, tmpf.name], check=True)
                with path.open(encoding='utf8') as fp:
                    new_data = json.load(fp)
                db.import_config(new_data)
                break
            except Exception:
                traceback.print_exc(file=sys.stdout)
                print()
                print('Errors occurred while editing configuration.')
                while True:
                    choice = input('(c)ancel, (e)dit again, (r)evert edits? ').lower()
                    if choice == 'c':
                        return
                    elif choice == 'e':
                        break
                    elif choice == 'r':
                        with path.open('w', encoding='utf8') as fp:
                            json.dump(data, fp, indent=4)
                        break

@command('config', 'edit', 'map', help='Edit mapping configuration with an external editor')
@arg('-e', '--editor', help='Editor command (defaults to $EDITOR environment variable)')
def edit_config_map(db:BackupDB, parser, args):
    edit_config(db, parser, args, ('path_map',))

@command('config', 'edit', 'path', help='Edit path configuration with an external editor')
@arg('vpath', nargs='?', metavar='VPATH', completer=complete_virtual_path, help='Virtual paths to edit')
@arg('-e', '--editor', help='Editor command (defaults to $EDITOR environment variable)')
def edit_config_path(db:BackupDB, parser, args):
    edit_config(db, parser, args, ('path_config', 'path_prefix'), canonical_path(args.vpath, True))

@command('config', 'edit', 'disks', help='Edit disk configuration with an external editor')
@arg('-e', '--editor', help='Editor command (defaults to $EDITOR environment variable)')
def edit_config_disk(db:BackupDB, parser, args):
    edit_config(db, parser, args, ('disks',))

##########################################################################################
# Debug
##########################################################################################

@command('debug', help='Debugging commands')
def cmd_debug(db:BackupDB, parser, args):
    parser.print_help()

@command('debug', 'hash', help='Hash a file')
@arg('path', type=Path, help='File to hash')
def cmd_debug_hash(db:BackupDB, parser, args):
    from .sourcefile import SourceFile
    srcf = SourceFile(args.path)
    print(f'hashing {args.path}')
    srcf.copy()
    print(f'hash = {srcf.hash}')

@command('debug', 'info', help='Show debug info about a path')
@arg('vpath', completer=complete_virtual_path, help='Virtual path to file')
def cmd_debug_info(db:BackupDB, parser, args):
    sql = '''
      SELECT
        ft.hash, ft.size, ft.last_modified, ft.ino, ft.metadata,
        ob.blocksize, ob.nexus, ob.refs, ob.copies
      FROM file_tree ft
      LEFT JOIN object ob ON ob.hash = ft.hash
      WHERE ft.virtual_path = ?
    '''

    args.vpath = canonical_path(args.vpath)
    row = db.query_one(sql, (args.vpath,))
    if not row:
        print(f'{args.vpath}: not found.')
        return

    hash, size, lastmod, ino, meta, blocksize, nexus, refs, copies = row
    mtimetxt = time.strftime('%Y-%m-%d %H:%M:%S%z', time.localtime(restore_mtime(lastmod)))

    print(f'{args.vpath}:')
    print(f'  Hash: {hash}')
    print(f'  Size: {addcomma(size)} ({addcomma(round_blocks(size))})')
    print(f'  Modified: {lastmod} ({mtimetxt})')
    print(f'  Inode: {ino}')
    if refs is not None:
        print(f'  Refs: {refs}')
    if nexus is not None:
        print(f'Nexus: {copies}, {nexus!r} ({" ".join(db.nexus_disks(nexus))})')
    if refs > 1:
        print('Identical files:')
        for nvpath, lastmod in db.execute('SELECT path, last_modified FROM object WHERE hash = ? AND path <> ?', (hash, args.vpath)):
            mtimetxt = time.strftime('%Y-%m-%d %H:%M:%S%z', time.localtime(restore_mtime(lastmod)))
            print(f'   ({mtimetxt}) {nvpath}')



import sys
import re
import json
import subprocess
import logging
import shutil

from typing import Optional
from dataclasses import dataclass
from collections import defaultdict
from pathlib import Path

import psutil

from .constants import DISK_CONF_NAME, OBJECT_DIR_NAME, LEGACY_OBJECT_DIR_NAME
from .utils import canonical_uuid, hash_relpath, read_file, write_file

log = logging.getLogger(__name__)

RX_HEX = re.compile(r'^[0-9a-fA-F]+$')

def find_mountpoint(path, partition_by_mountpoint):
    path = path.resolve()
    cpath = path
    cstat = path.stat()
    while True:
        partition = partition_by_mountpoint.get(str(cpath))
        if partition:
            break
        ppath = cpath.parent
        pstat = ppath.stat()
        if cpath == ppath or cstat.st_dev != pstat.st_dev:
            break
        cpath = ppath
        cstat = pstat

    return cpath, str(path.relative_to(cpath)), shutil.disk_usage(path).total, partition

@dataclass
class Partition:
    device: str
    mountpoint: Path
    size: int = 0
    label: Optional[str] = None
    fstype: Optional[str] = None
    fsuuid: Optional[str] = None

    @classmethod
    def list(cls):
        rtn = []
        if sys.platform == 'linux':
            try:
                p = subprocess.run(
                    'lsblk --json -p -b -o name,size,label,uuid,mountpoint,fstype,fsavail,fssize --list'.split(),
                    encoding='ascii', stdout=subprocess.PIPE, check=True
                )
                jsdata = json.loads(p.stdout)
                for dev in jsdata['blockdevices']:
                    mountpoint = dev.get('mountpoint')
                    part = cls(
                        device=dev.get('name'),
                        mountpoint=Path(mountpoint) if mountpoint else None,
                        size=dev.get('fssize') or dev.get('size'),
                        label=dev.get('label'),
                        fstype=dev.get('fstype'),
                        fsuuid=dev.get('uuid'),
                    )
                    if part.mountpoint or part.fsuuid:
                        rtn.append(part)

                return rtn

            except (OSError, ValueError, KeyError) as e:
                log.warning(f'Unable to use `lsblk`: {e}', file=sys.stderr)

        for disk in psutil.disk_partitions():
            if not disk.mountpoint:
                continue

            rtn.append(cls(device=disk.device, mountpoint=Path(disk.mountpoint), fstype=disk.fstype))
        return rtn

@dataclass
class BackupDisk:
    uuid: str
    name: str
    nexus_index: int = 0
    relative_path: Optional[str] = None
    size: int = 0
    fstype: Optional[str] = None
    fsuuid: Optional[str] = None
    used_space: int = 0
    mountpoint: Optional[Path] = None
    data_path: Optional[Path] = None
    device_name: Optional[str] = None

    def write_config(self):
        jsdata = json.dumps({
            "uuid": self.uuid,
            "name": self.name
        }, indent=4)
        write_file(self.data_path / DISK_CONF_NAME, jsdata)

    @classmethod
    def read_config(cls, path):
        try:
            with (path / DISK_CONF_NAME).open('r') as fp:
                jsdata = json.load(fp)

            uuid = canonical_uuid(jsdata['uuid'])
            name = jsdata.get('name')
            return cls(uuid, name, data_path=path)

        except FileNotFoundError:
            # Legacy configuration
            uuid = read_file(path / '.distbackup.uuid')
            if not uuid:
                return None
            name = read_file(path / '.distbackup.name')
            rdisk = cls(uuid, name, data_path=path)
            rdisk.write_config()
            return cls(uuid, name, data_path=path)

        except (ValueError, OSError, KeyError):
            return None

    def set_from_path(self, path: Path):
        partitions = Partition.list()
        partition_by_mountpoint = {part.mountpoint: part for part in partitions if part.mountpoint}
        mountpt, relpath, totalsize, partition = find_mountpoint(path, partition_by_mountpoint)

        if not self.size:
            # Reserve 5% of the disk, or 10G, whichever is smaller
            reserve = min(10_000_000_000, totalsize // 20)
            self.size = (totalsize - reserve) & ~0xFFF

        if partition is not None:
            self.fstype = partition.fstype
            self.fsuuid = partition.fsuuid

        self.mountpoint = mountpt
        self.data_path = path
        self.relative_path = relpath

    def hashpath(self, hash):
        return self.data_path / hash_relpath(hash)

    def migrate_data(self):
        if not self.data_path:
            return

        object_path = self.data_path / OBJECT_DIR_NAME
        if not object_path.exists():
            legacy_path = self.data_path / LEGACY_OBJECT_DIR_NAME
            if legacy_path.exists():
                legacy_path.rename(object_path)
            else:
                object_path.mkdir(parents=True, exist_ok=True)

    def walk_data(self):
        base = self.data_path / OBJECT_DIR_NAME

        for subdir in sorted(base.iterdir()):
            if not RX_HEX.match(subdir.name):
                continue

            for hashf in sorted(subdir.iterdir()):
                hash = hashf.name
                if not RX_HEX.match(hash):
                    continue

                try:
                    stat = hashf.lstat()
                except OSError:
                    continue

                yield hash, hashf, stat

    @classmethod
    def find_disk_data_paths(cls, disks):
        disk_by_fsuuid = defaultdict(list)
        disk_by_relpath = defaultdict(dict)
        for disk in disks:
            if disk.relative_path and disk.data_path is None:
                if disk.fsuuid:
                    disk_by_fsuuid[disk.fsuuid].append(disk)
                else:
                    disk_by_relpath[disk.relative_path][disk.uuid] = disk

        for part in Partition.list():
            if not part.mountpoint:
                if part.fsuuid:
                    for disk in disk_by_fsuuid.get(part.fsuuid, ()):
                        disk.device_name = part.device
                continue

            if part.fsuuid:
                for disk in disk_by_fsuuid.get(part.fsuuid, ()):
                    rdisk = BackupDisk.read_config(part.mountpoint / disk.relative_path)
                    if rdisk:
                        if rdisk.uuid != disk.uuid:
                            log.warning(f'Disk on partition {part.fsuuid} ({rdisk.data_path}) does not match expected UUID!', file=sys.stderr)
                            log.warning(f'  Expected: {disk.name} {disk.uuid}')
                            log.warning(f'  Got     : {rdisk.name} {rdisk.uuid}')
                            continue

                        disk.mountpoint = part.mountpoint
                        disk.data_path = rdisk.data_path
                        disk.device_name = part.device
                        log.debug(f'Found path for {disk.name} using partition UUID: {disk.data_path}')

                        if disk.name != rdisk.name:
                            disk.write_config()

            for path, disk_map in disk_by_relpath.items():
                # If we've already found all disks for this path, don't try to read it
                if not disk_map:
                    continue

                if not any(disk for disk in disk_map.values() if (disk.fstype is None or part.fstype == disk.fstype)):
                    continue

                rdisk = BackupDisk.read_config(part.mountpoint / path)
                if rdisk:
                    disk = disk_map.pop(rdisk.uuid, None)
                    if disk is None:
                        continue

                    disk.mountpoint = part.mountpoint
                    disk.data_path = rdisk.data_path
                    disk.device_name = part.device
                    log.debug(f'Found path for {disk.name} by scanning: {disk.data_path}')

                    if disk.name != rdisk.name:
                        disk.write_config()



import re
import time
import logging
from pathlib import Path

# This will remove all the fancy colors that distbackup works so hard to add to its console output :'(
RX_UNSTYLE = re.compile(r'\033\[[;0-9]*[a-zA-Z]')

class FormatDelegator(logging.Formatter):
    '''A formatter which can delegate to other formatters based on log level'''
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._level_delegates = {}

    def set_level_formatter(self, level:int, formatter:logging.Formatter):
        self._level_delegates[level] = formatter

    def set_level_format(self, level:int, fmt:str, *a, **kw):
        self.set_level_formatter(level, logging.Formatter(fmt, *a, **kw))

    def format(self, record):
        formatter = self._level_delegates.get(record.levelno, None)
        if formatter:
            return formatter.format(record)
        else:
            return super().format(record)

class FilePatternLogger(logging.Handler):
    def __init__(self, filename_pattern=''):
        super().__init__()
        self.current_file = None
        self.current_log_path = None
        self.closed = False

        # Convert from pathlib.Path to string if necessary
        filename_pattern = str(filename_pattern)
        self.filename_pattern = filename_pattern
        self.bydate = bool(filename_pattern and '@@' in filename_pattern)

        # If we're not splitting based on date, just open the output now.
        if filename_pattern and not self.bydate:
            self._open(filename_pattern)

    def close(self):
        if self.current_file is not None:
            self.current_file.close()
            self.current_file = None
        self.closed = True

    def _open(self, path):
        # Don't reopen after close
        if not self.closed:
            path = Path(path)
            path.parent.mkdir(parents=True, exist_ok=True)
            self.current_file = Path(path).open('a', encoding='utf8')

    def emit(self, record):
        if self.closed:
            return

        text = RX_UNSTYLE.sub('', self.format(record))
        localtime = time.localtime(record.created)
        if self.bydate:
            cdate = time.strftime("%Y-%m-%d", localtime)
            current_log_path = self.filename_pattern.replace('@@', cdate)
            if current_log_path != self.current_log_path:
                if self.current_file:
                    self.current_file.close()
                self.current_log_path = current_log_path
                self._open(current_log_path)

        if self.current_file:
            self.current_file.write(text + '\n')
            self.current_file.flush()



import statistics
from dataclasses import dataclass, field

from .utils import disk_in_nexus

@dataclass
class Nexus:
    nexusid: str
    level: int = 0
    refs: int = 0
    totalsize: int = 0
    saturated_size: int = 0
    disks: str = ''

@dataclass
class NexusLevel:
    level: int
    members: list[Nexus] = field(default_factory=list)
    totalsize: int = 0
    saturated_size: int = 0
    imbalance: float = 0.0
    imbalance_change_cache: dict = field(default_factory=dict)

    def add_member(self, m):
        self.totalsize += m.totalsize
        self.saturated_size += m.saturated_size
        self.members.append(m)

    def change_size(self, nexus, diff, saturated:bool=False):
        self.totalsize += diff
        nexus.totalsize += diff
        if saturated:
            self.saturated_size += diff
            nexus.saturated_size += diff

        self.imbalance_change_cache.clear()
        self.update_imbalance()

    def calc_imbalance(self, modified_nexus=None, size_diff_order=0, sign=1):
        '''Calculate how "balanced" the member nexuses would be with the given change'''
        res = self.imbalance_change_cache.get(sign * size_diff_order)
        if res is not None:
            return res

        size_diff = sign * (1 << size_diff_order)
        res = 0.0
        if len(self.members) > 1:
            mean = (self.totalsize + size_diff) / len(self.members)
            res = statistics.stdev([(nexus.totalsize + size_diff) if nexus is modified_nexus else nexus.totalsize for nexus in self.members], mean)

        self.imbalance_change_cache[sign * size_diff_order] = res

        return res

    def update_imbalance(self):
        if len(self.members) > 1:
            mean = self.totalsize / len(self.members)
            self.imbalance = statistics.stdev([nexus.totalsize for nexus in self.members], mean)
        else:
            self.imbalance = 0.0

class NexusLevelDict(dict):
    def __init__(self, nexus_seq):
        super().__init__()
        if isinstance(nexus_seq, dict):
            nexus_seq = nexus_seq.values()

        for nexus in nexus_seq:
            self[nexus.level].add_member(nexus)

    def __missing__(self, key):
        r = self[key] = NexusLevel(key)
        return r

    def move_object(self, old_nexus, new_nexus, size, maxcopies=None):
        self[old_nexus.level].change_size(old_nexus, -size, maxcopies is not None and old_nexus.level >= maxcopies)
        self[new_nexus.level].change_size(new_nexus, size, maxcopies is not None and new_nexus.level >= maxcopies)

    def calc_imbalance_change(self, from_nexus, to_nexus, size_diff_order):
        from_level = self[from_nexus.level]
        to_level = self[to_nexus.level]

        imbalance_change_from = from_level.calc_imbalance(from_nexus, size_diff_order, -1) - from_level.imbalance
        imbalance_change_to = to_level.calc_imbalance(to_nexus, size_diff_order, 1) - to_level.imbalance

        # Weight the imbalance change to higher-level nexus more.
        if from_nexus.level > to_nexus.level:
            imbalance_change_from *= 2
        else:
            imbalance_change_to *= 2

        return imbalance_change_to + imbalance_change_from

    def get_copy_data(self, nexus_index=None):
        copy_data = [(0, 0)] * (max(self) + 1)

        if nexus_index is not None:
            for index, level in self.items():
                all_nexus = [nexus for nexus in level.members if disk_in_nexus(nexus.nexusid, nexus_index)]
                copy_data[index] = (
                    sum(nexus.totalsize for nexus in all_nexus),
                    sum(nexus.saturated_size for nexus in all_nexus)
                )
        else:
            for index, level in self.items():
                copy_data[index] = (level.totalsize, level.saturated_size)

        return copy_data



import os
import time
import logging

from collections import deque, Counter

from ktpanda.textcolor import Colorizer

from .errors import ArgumentError, IncompleteUpdateError
from .utils import (CacheDict, addcomma, check_dir, int_mtime, nexus_with_disk, nexus_without_disk,
                    prettysize, prettysize_clr, restore_mtime, try_unlink, disk_in_nexus)
from .nexus import Nexus, NexusLevel, NexusLevelDict
from .objects import ObjectCopyEntry, ObjectGroup
from .database import BackupDB
from .sourcefile import SourceFile

log = logging.getLogger(__name__)

class Backup:
    def __init__(self, db:BackupDB, sync=False):
        self.db:BackupDB = db
        self.temp_index:int = 1
        self.sync:bool = sync

    def tempname(self):
        idx = self.temp_index
        self.temp_index = idx + 1
        path = self.db.current_disk.data_path / f'distbackup-tmp.{os.getpid()}.{idx}'
        return path

    def print_delete(self, del_obj):
        log.info(f"delete ({prettysize(del_obj.size)}) ({del_obj.nexusto.disks}) {del_obj.path}")

    def copy_object(self, copy_obj, copyremain=None):
        '''Copy an object to the current backup disk.'''
        srcf = None
        vpath = None
        with self.db.execute('SELECT virtual_path, last_modified FROM file_tree WHERE hash = ?', (copy_obj.hash,)) as curs:
            for path, lastmod in curs:
                fpath = self.db.get_file_path(path)
                if fpath is None:
                    log.warning(f"Could not find real file for virtual path {path}")
                    continue

                try:
                    st = fpath.stat()
                    srcf = SourceFile(fpath, st)

                    if int_mtime(st.st_mtime) != lastmod:
                        raise OSError()

                    vpath = path
                    break
                except OSError as e:
                    log.warning(f"source modified ({path}): {e}")

        if not vpath:
            return False

        txt = Colorizer().t(f'copy {vpath} (').t(prettysize_clr(copy_obj.size)).t(f') ({copy_obj.nexusfrom.disks})')

        if copyremain is not None:
            txt.t(' limit ').t(prettysize_clr(copyremain))

        log.info(txt.get())

        try:
            dstfn = self.tempname()
            srcf.setdest(dstfn)
            srcf.copy(self.db.check_run_deferred, self.sync)
            if srcf.hash != copy_obj.hash:
                log.info(f'WARNING: data for {vpath} does not match hash: expected {copy_obj.hash}, got {srcf.hash}')
                # Set the file's hash to the actual value. It will get picked up on the next update/copy.
                self.db.defer(self.db.set_file_hash, vpath, srcf.hash)
                return False
            destname = self.db.current_disk.hashpath(copy_obj.hash)
            check_dir(destname)
            dstfn.replace(destname)

            self.db.defer(self.db.set_object_on_disk, copy_obj.hash, self.db.current_disk.nexus_index, True)

            dstfn = None

        finally:
            if dstfn:
                try_unlink(dstfn)

        return True

    def delete_objects(self, to_delete, printf=None):
        for del_obj in to_delete:
            if printf:
                printf(del_obj)
            self.db.check_run_deferred()
            if try_unlink(self.db.current_disk.hashpath(del_obj.hash)):
                self.db.defer(self.db.set_object_on_disk, del_obj.hash, self.db.current_disk.nexus_index, False)
            else:
                log.info(f'warning: could not delete hash {hash}')
                continue

    def backup_files(self, simulate=False, noflush=False, limit_bytes=None, limit_copies=None):
        try:
            return self._backup_files(simulate, noflush, limit_bytes, limit_copies)
        finally:
            self.db.check_run_deferred(True)

    def _backup_files(self, simulate=False, noflush=False, limit_bytes=None, limit_copies=None):
        self.db.require_disk()
        if not self.db.current_disk.data_path and not simulate:
            raise ArgumentError('updating target requires a backup destination or --simulate')

        missing_hash_count = self.db.query_scalar("SELECT count(*) FROM file_tree WHERE hash = ''")
        if missing_hash_count:
            raise IncompleteUpdateError(f'Source update incomplete: {missing_hash_count} files are missing hashes')

        if not simulate:
            self.db.cleanrefs()
            self.db.copy_database(self.db.current_disk.data_path)
            self.db.current_disk.migrate_data()

        # nexus_dict: Maps rows from the `nexus` table to `Nexus` objects
        nexus_dict: CacheDict[str, Nexus] = CacheDict(self.db.create_empty_nexus)
        nexus_dict.update(self.db.load_nexus())

        copyremain = limit_bytes
        self.db.update_free_space()

        total_copied = Counter()
        total_deleted = Counter()

        nexus_index = self.db.current_disk.nexus_index

        # objects_off_disk: Contains all objects not present on the disk. We will pick as
        # many objects from this container to copy to the disk. It starts out as a
        # CacheDict so we can group objects together by priority. Later, we convert it
        # into a list of groups sorted by priority.
        objects_off_disk = CacheDict(ObjectGroup)

        # objects_on_disk: Contains all objects on the disk that we can delete.
        objects_on_disk = CacheDict(ObjectGroup)

        available_space = self.db.current_disk.size - self.db.current_disk.used_space

        to_delete = []

        # Search through all objects that have references from file_tree.
        with self.db.execute('SELECT hash, size, blocksize, size_order, copies, refs, nexus, priority, maxcopies, last_path, last_modtime FROM object') as curs:
            for hash, size, blocksize, size_order, copies, refs, current_nexus, priority, maxcopies, last_path, last_modtime in curs:
                # If the current disk is in the object's nexus, that means the object is
                # currently on the disk, and we can delete it to make room for higher priority
                # objects,
                entry = ObjectCopyEntry(
                    hash=hash, size=size, blocksize=blocksize, size_order=size_order, refs=refs, groupkey=None,
                    nexusfrom=nexus_dict[current_nexus], maxcopies=maxcopies, path=last_path, modtime=last_modtime
                )

                if disk_in_nexus(current_nexus, nexus_index):
                    entry.groupkey = copies, -priority
                    entry.nexusto = nexus_dict[nexus_without_disk(current_nexus, nexus_index)]
                    # Immediately delete any objects with no more references or have too many copies
                    if (maxcopies is not None and copies > maxcopies) or refs == 0:
                        to_delete.append(entry)
                        continue

                    # Never delete files that have only one copy, regardless of priority.
                    if copies <= 1:
                        continue

                    objects_on_disk[entry.groupkey].add_entry(entry)
                else:
                    entry.groupkey = copies + 1, -priority
                    entry.nexusto = nexus_dict[nexus_with_disk(current_nexus, nexus_index)]

                    # Never copy files that already have the maximum number of copies, and
                    # don't copy objects that have no references.
                    if (maxcopies is not None and copies >= maxcopies) or refs == 0:
                        continue

                    # Don't copy files that already have the maximum number of copies for
                    # this session.
                    if limit_copies is not None and copies >= limit_copies:
                        continue

                    objects_off_disk[entry.groupkey].add_entry(entry)

        # levels: Maps `level` -> NexusLevel, which groups all Nexus objects by count of
        # disks in the set.
        levels: NexusLevelDict[int, NexusLevel] = NexusLevelDict(nexus_dict)
        nexus_dict.clear()

        # Delete all the dead and excess objects
        for del_obj in to_delete:
            timestamp = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(restore_mtime(del_obj.modtime)))
            log.info(f"delete {del_obj.hash} ({timestamp}): {del_obj.path})")

            # Entries with refs == 0 are already excluded from size counts
            if del_obj.refs > 0:
                levels.move_object(del_obj.nexusfrom, del_obj.nexusto, del_obj.blocksize, del_obj.maxcopies)
                available_space += del_obj.blocksize
                total_deleted[del_obj.nexusfrom.level] += del_obj.blocksize

        if not simulate:
            self.delete_objects(to_delete)

        # Sort the groups so that the first element is the one with the lowest number of
        # copies and highest proiority.
        objects_off_disk = deque(sorted(objects_off_disk.values()))
        objects_on_disk = deque(sorted(objects_on_disk.values()))

        # All the objects that we could not copy, either because they were either
        # modified, moved, or deleted since the last update.
        unable_to_copy = []


        # Calculate the initial imbalance value for each level.
        for grp in levels.values():
            grp.update_imbalance()

        if self.db.verbose:
            log.info(f'Imbalance before ({len(levels)}):')
            for level, grp in sorted(levels.items()):
                log.info(f'  Level {level:2}: level={len(grp.members):3d} size={addcomma(grp.totalsize):>20} imbalance={addcomma(format(grp.imbalance, ".2f")):>24}')

        if self.db.verbose >= 2:
            for grp in objects_on_disk:
                log.info(f'reclaimable space for {grp.groupkey} = {addcomma(grp.totalsize)}')

        # Main backup loop
        while True:
            self.db.check_run_deferred()
            try:
                # Grab the highest priority group from objects_off_disk, removing any that are empty.
                while len((src_group := objects_off_disk[0]).objects) == 0:
                    objects_off_disk.popleft()
            except IndexError:
                # IndexError means there are no more groups and therefore no more objects to copy.
                break

            # Remove any groups from the delete set whose priority is greater than or
            # equal to the priority of the group we're copying from. An object can
            # only replace an object of a lower priority.
            try:
                while objects_on_disk[0].groupkey <= src_group.groupkey:
                    objects_on_disk.popleft()

                while len(objects_on_disk[-1].objects) == 0:
                    objects_on_disk.pop()
            except IndexError:
                pass

            # Calculate the maximum size of an object we can copy.
            if noflush:
                reclaimable_space = 0
            else:
                reclaimable_space = sum(grp.totalsize for grp in objects_on_disk)

            total_available = available_space + reclaimable_space

            # If we have a copy limit, cap `total_available`
            if copyremain is not None and total_available > copyremain:
                total_available = copyremain

            # Calculate the `imbalance_change` value for all the objects in this group if
            # it has not been calculated before.
            if not src_group.sorted:
                src_group.sort(levels)

            # Grab a random object to copy.
            try:
                copy_obj = src_group.pop_entry(total_available)
            except IndexError:
                # IndexError means no more items from this group could be copied. Move on
                # to the next group.
                continue

            assert not disk_in_nexus(copy_obj.nexusfrom.nexusid, nexus_index), f'copy_obj already on disk {copy_obj}'

            if self.db.verbose >= 2:
                log.info(f'trying to copy {copy_obj.path} ({addcomma(copy_obj.size)}), avail_space={addcomma(available_space)}')

            # We may still fail to copy, but for now, pretend we can. If we don't copy,
            # these changes will be rolled back. This is only for the current process, no
            # changes are made to the database unless files are actually copied or
            # deleted.
            if copyremain is not None:
                copyremain -= copy_obj.blocksize
            available_space -= copy_obj.blocksize
            levels.move_object(copy_obj.nexusfrom, copy_obj.nexusto, copy_obj.blocksize, copy_obj.maxcopies)

            to_delete = []

            # Initialize the group we're looking for to the end of objects_on_disk. We
            # want to avoid deleting empty sets because we may revert,
            del_group_index = len(objects_on_disk) - 1
            if available_space < 0:
                if self.db.verbose >= 2:
                    for grp in objects_on_disk:
                        log.info(f'reclaimable space for {grp.groupkey} = {addcomma(grp.totalsize)}')
                    log.info(f'total reclaimable space = {addcomma(reclaimable_space)}')

                if reclaimable_space < -available_space:
                    continue

            while available_space < 0 and not noflush:
                del_group = None

                while del_group_index >= 0 and len((del_group := objects_on_disk[del_group_index]).objects) == 0:
                    # Find the lowest priority group, skipping ones that are empty.
                    del_group_index -= 1

                if del_group is None or del_group_index < 0:
                    # No more files that can be deleted.
                    break

                # Calculate the `imbalance_change` value for all the objects in this group
                # if it has not been calculated before.
                if not del_group.sorted:
                    del_group.sort(levels)

                # Grab a random object to delete.
                del_obj = del_group.pop_entry()

                assert disk_in_nexus(del_obj.nexusfrom.nexusid, nexus_index), f'del_obj not on disk {del_obj}'

                to_delete.append(del_obj)
                levels.move_object(del_obj.nexusfrom, del_obj.nexusto, del_obj.blocksize, del_obj.maxcopies)
                available_space += del_obj.blocksize
                if self.db.verbose >= 2:
                    log.info(f'del {del_obj.path} ({addcomma(del_obj.size)}), avail_space={addcomma(available_space)}')

            copied = False
            # If we have enough space, then copy the object.
            if available_space >= 0:
                for del_obj in to_delete:
                    total_deleted[del_obj.nexusfrom.level] += del_obj.blocksize

                if simulate:
                    for obj in to_delete:
                        log.info(f'would delete {obj.path} {prettysize(obj.size)} ({obj.nexusfrom.disks})')
                    log.info(f'would copy {copy_obj.path} {prettysize(copy_obj.size)} ({copy_obj.nexusfrom.disks})')
                    copied = True
                else:
                    self.delete_objects(to_delete, self.print_delete)
                    if self.copy_object(copy_obj, copyremain):
                        copied = True
                    else:
                        unable_to_copy.append(copy_obj)
            else:
                if self.db.verbose >= 2:
                    log.info(f'Not enough space to copy {obj.path}')

            # If we didn't actually copy the file, roll back the changes that we made.
            if copied:
                total_copied[copy_obj.nexusto.level] += copy_obj.blocksize
                for grp in objects_on_disk:
                    grp.commit()
            else:
                if copyremain is not None:
                    copyremain += copy_obj.blocksize
                available_space += copy_obj.blocksize
                levels.move_object(copy_obj.nexusto, copy_obj.nexusfrom, copy_obj.blocksize, copy_obj.maxcopies)
                for grp in objects_on_disk:
                    if self.db.verbose >= 2:
                        if grp.saved_objects is not None:
                            log.info(f'roll back delete group {grp.groupkey} {len(grp.objects)} -> {len(grp.saved_objects)}')
                    grp.rollback()

        log.info(f'Copied {prettysize(sum(total_copied.values()))}:')
        for level, bytes in sorted(total_copied.items()):
            log.info(f'  Level {level - 1} -> {level}: {prettysize(bytes)}')

        log.info(f'Deleted {prettysize(sum(total_deleted.values()))}:')
        for level, bytes in sorted(total_deleted.items()):
            log.info(f'  Level {level} -> {level - 1}: {prettysize(bytes)}')

        if unable_to_copy:
            log.info('some files could not be copied (recommend --update)')

        if not simulate:
            self.db.cleanrefs()
            self.db.copy_database()
        return levels



#!/usr/bin/python3
import sys
import os
import logging
from pathlib import Path

try:
    import argcomplete
except ImportError:
    argcomplete = None

from .errors import ArgumentError
from .logger import FormatDelegator, FilePatternLogger
from .cli import open_db
from .utils import COMMANDS, Progress, ProgressBar

def init_arguments():
    root_parser = COMMANDS[()].parser

    db_path = os.getenv('DISTBACKUP_PATH')
    if db_path is not None:
        default_db_path = Path(db_path)
    else:
        configdir = os.getenv('XDG_CONFIG_HOME')
        if configdir:
            configdir = Path(configdir)
        else:
            configdir = Path.home() / '.config'

        default_db_path = configdir / 'distbackup'


    root_parser.set_defaults(
        command_name=(),
        sync=False,
        nocopy=False,
        no_hashcopy=False,
    )

    root_parser.add_argument(
        '-d', '--db', type=Path, metavar='PATH',
        default=default_db_path,
        help='Location of primary database (default: %(default)s)')

    root_parser.add_argument(
        '-l', '--log', action='store',
        help='Write logs to a file. If the pattern contains "@@", it is replaced '
        'with the current date. (default: <dbpath>/log/log-@@.txt)'
    )

    root_parser.add_argument(
        '-L', '--nolog', action='store_true',
        help='Do not write to log file.'
    )

    root_parser.add_argument(
        '-v', '--verbose', action='count',
        default=0,
        help='Verbose logging'
    )

    root_parser.add_argument(
        '-e', '--explain', action='store_true',
        help='Explain queries as they are executed'
    )

    root_parser.add_argument(
        '-n', '--noaction', action='store_true',
        help='Do not actually perform any actions'
    )

    if argcomplete:
        argcomplete.autocomplete(root_parser, validator=lambda current_input, keyword_to_check_against: True)

    return root_parser


def main():
    parser = init_arguments()
    args = parser.parse_args()

    Progress.set_global(ProgressBar())

    if args.no_hashcopy:
        from .sourcefile import SourceFile
        SourceFile.ENABLE_HASHCOPY = False


    root = logging.getLogger()
    root.setLevel(logging.DEBUG if args.verbose else logging.INFO)

    console_formatter = FormatDelegator('%(levelname)s: %(message)s')
    console_formatter.set_level_format(logging.INFO, '%(message)s')

    log_formatter = logging.Formatter('%(asctime)s %(levelname)s: [%(name)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(console_formatter)
    root.addHandler(console_handler)

    if args.nolog or args.log == '':
        logpath = None
    elif not args.log:
        dbpath = args.db
        if dbpath.is_dir():
            dbdir = dbpath
        else:
            dbdir = dbpath.parent
        logpath = dbdir / 'log' / 'log-@@.txt'
    else:
        logpath = args.log

    if logpath:
        logfile_handler = FilePatternLogger(logpath)
        logfile_handler.setFormatter(log_formatter)
        root.addHandler(logfile_handler)

    log = logging.getLogger('main')

    if args.verbose:
        log.info(f'distbackup starting: {sys.argv[1:]!r}')

    db = open_db(args)

    cmd = COMMANDS[args.command_name]
    try:
        cmd.func(db, cmd.parser, args)
    except ArgumentError as e:
        log.error(str(e))
        return
    except KeyboardInterrupt:
        log.info('Interrupted')
    except Exception:
        log.exception(f'Exception running {" ".join(args.command_name)}')

if __name__ == '__main__':
    main()



import sys
import re
import time
import os
import json
import logging

from hashlib import sha256

try:
    import hashcopy
except ImportError:
    hashcopy = None

from .constants import META_SUFFIX
from .utils import round_blocks, progress
from .errors import FileChanged

log = logging.getLogger(__name__)

class SourceFile:
    ENABLE_HASHCOPY = True

    def __init__(self, path, stat=None):
        self.path = path
        self.size = 0
        self.blocksize = 0
        self.mode = 0
        self.mtime = 0
        self.inode = 0

        self.fp = None
        self.dest = None
        self.destfp = None
        self.hash = None
        self.priority = None
        self.maxcopies = None

        self.metadata = {}

        if stat is None:
            stat = path.lstat()
        self.setstat(stat)

    def read_meta(self):
        try:
            metapath = self.path.with_name(self.path.name + META_SUFFIX)
            with metapath.open() as fp:
                self.metadata = json.load(fp)

            hash = self.metadata.get('hash')
            if hash:
                if not (isinstance(hash, str) and re.match(r'^[0-9a-f]{64}$', hash, re.I)):
                    raise ValueError(f'Invalid hash: {hash}')
                self.hash = hash.lower()

            priority = self.metadata.get('priority')
            if priority is not None:
                if not (isinstance(priority, int) and priority >= 0):
                    raise ValueError(f'Invalid priority: {priority}')
                self.priority = priority

            maxcopies = self.metadata.get('maxcopies')
            if maxcopies is not None:
                if not (maxcopies == 'inf' or (isinstance(maxcopies, int) and maxcopies >= 1)):
                    raise ValueError(f'Invalid maxcopies: {maxcopies}')
                self.maxcopies = None if maxcopies == 'inf' else maxcopies
        except FileNotFoundError:
            pass
        except ValueError as e:
            print(f'WARNING: Metadata file {metapath} is corrupt: {e}', file=sys.stderr)


    def check_change(self):
        nst = self.path.lstat()
        if nst.st_mode != self.mode or nst.st_mtime != self.mtime or nst.st_size != self.size:
            self.setstat(nst)
            raise FileChanged()

    def _open(self):
        if self.fp is None:
            self.fp = self.path.open('rb')

    def reset(self):
        if self.fp is not None:
            self.fp.close()
            self.fp = None

    def setdest(self, dest):
        self._open()
        self.dest = dest
        self.destfp = dest.open('wb')

    def setstat(self, stat):
        self.size = stat.st_size
        self.blocksize = round_blocks(self.size)
        self.mode = stat.st_mode
        self.mtime = stat.st_mtime
        self.inode = stat.st_ino

    def copy(self, periodic=None, sync=False):
        self._open()
        ltime = time.time()
        pos = 0
        prog = False
        if hashcopy and SourceFile.ENABLE_HASHCOPY:
            hh = hashcopy.HashCopier(self.fp.fileno(), self.destfp.fileno() if self.destfp else -1)
        else:
            if SourceFile.ENABLE_HASHCOPY:
                log.warning('Hash accelerator is enabled but `hashcopy` module could not be imported')
                SourceFile.ENABLE_HASHCOPY = False
            hh = None
            sha = sha256()
        try:
            while True:
                if hh:
                    copylen = hh.update()
                else:
                    dat = self.fp.read(524288)
                    copylen = len(dat)
                    if self.destfp:
                        self.destfp.write(dat)

                if copylen == 0:
                    break

                pos += copylen
                ctime = time.time()
                if ctime >= ltime + 0.2:
                    if periodic:
                        periodic()
                    ltime = ctime
                    prog = True
                    pct = (pos * 100.0 / self.size)
                    progress(pct)

                # If we're using hash_helper, sync the destination after every
                # write. We're depending on readahead to fill the memory mapped buffer
                # while the sync is occurring.
                if hh:
                    if self.destfp and sync:
                        os.fsync(self.destfp.fileno())
                else:
                    sha.update(dat)

            self.reset()
            self.size = pos

            if self.destfp:
                if prog:
                    progress(100)
                if sync:
                    os.fsync(self.destfp.fileno())
                self.destfp.close()

            if hh:
                self.hash = hh.finalize().hex()
            else:
                self.hash = sha.hexdigest()

            return self.hash
        finally:
            if prog:
                progress(None)



import traceback
import time
import sqlite3
import uuid
import json
import logging

from typing import Optional
from pathlib import Path

# Third party imports
from ktpanda.sqlite_helper import SQLiteDB, in_transaction

from .constants import DBNAME, DEFAULT_CONFIG
from .disk import BackupDisk
from .nexus import Nexus
from .utils import (SQL_FUNCS, addcomma, canonical_path, canonical_uuid, compact_json,
                    disk_in_nexus, new_uuid, nexus_level, nexus_with_disk, nexus_without_disk,
                    pattern_to_regex, prefix_sql, Progress)
from .errors import ArgumentError, DiskNotFoundError
from .schema import PRE_DATA, SCHEMA_VERSION, TABLES_INDICES, TRIGGERS

log = logging.getLogger(__name__)

class BackupDB(SQLiteDB):
    common_schema = TABLES_INDICES + TRIGGERS + PRE_DATA
    schema_version = SCHEMA_VERSION

    def __init__(self, dbpath:Path, readonly:bool=False, verbose:bool=False):
        if dbpath.is_dir():
            dbpath = dbpath / DBNAME

        super().__init__(dbpath, readonly=readonly)

        self.verbose = verbose

        self.disk_by_index:dict[int, BackupDisk] = {}
        self.disk_by_uuid:dict[str, BackupDisk] = {}
        self.disk_by_name:dict[str, BackupDisk] = {}
        self._disk_name_index:list[tuple[str, int]] = []

        self.current_disk:Optional[BackupDisk] = None
        self.defer_queue:list[tuple] = []
        self.last_run_deferred:float = 0

        self.path_map:dict[str, Path] = {}
        self.path_cache:dict[str, Path] = {}
        #self.diropt_cache = {}

        self.nocopydb:bool = False

    ##########################################################################################
    # Upgrades
    ##########################################################################################

    def _upgrade_to_13(self, oldvers):
        self.backend.execute('ALTER TABLE disks ADD COLUMN relative_path TEXT')

    def _upgrade_to_16(self, oldvers):
        self.backend.execute("ALTER TABLE file_names ADD COLUMN metadata TEXT NOT NULL DEFAULT ''")

    def _upgrade_to_18(self, oldvers):
        for column in ('file_names.path', 'path_map.virtual_path', 'path_config.virtual_path', 'links.path', 'files.last_path'):
            table, column = column.split('.')
            sql = f"UPDATE {table} SET {column} = '/' || {column} WHERE {column} NOT LIKE '/%'"
            self.backend.execute(sql)

    def _upgrade_to_20(self, oldvers):
        if self.backend.execute("SELECT name FROM sqlite_master WHERE name = 'source_path'"):
            self.backend.execute('ALTER TABLE source_path RENAME TO path_map')

        if self.backend.execute("SELECT name FROM sqlite_master WHERE name = 'source_config'"):
            self.backend.execute('ALTER TABLE source_config RENAME TO path_config')

    def _drop_triggers(self):
        triggers = [row[0] for row in self.backend.execute("SELECT name FROM sqlite_master WHERE type = 'trigger'")]
        for trigger in triggers:
            self.backend.execute(f'DROP TRIGGER {trigger}')

    def _upgrade_to_21(self, oldvers):
        '''*BIG UPDATE* - drop all triggers, create the new tables, copy
        the data over, then let it create the new triggers'''

        self._drop_triggers()

        self.exec_schema(TABLES_INDICES)

        self.backend.create_function('newid', 0, new_uuid)
        self.backend.execute('INSERT INTO object(hash, size, nexus, refs, last_path, last_modtime) SELECT hash, size, nexus, refs, last_path, last_lastmod FROM files')
        self.backend.execute('INSERT INTO file_tree(virtual_path, last_modified, hash, ino, size, metadata) SELECT path, lastmod, hash, ino, size, metadata FROM file_names')
        self.backend.execute('INSERT INTO symbolic_link(virtual_path, target) SELECT path, dest FROM links')
        self.backend.execute('INSERT INTO disk(uuid, name, nexus_index, relative_path, size) SELECT IFNULL(uuid, newid()), name, id - 1, relative_path, size FROM disks')
        self.backend.execute('INSERT INTO nexus(nexus, refs, totalsize) SELECT nexus, refs, totalsize FROM dist_nexus')
        self.backend.execute('DROP TABLE files')
        self.backend.execute('DROP TABLE file_names')
        self.backend.execute('DROP TABLE links')
        self.backend.execute('DROP TABLE disks')
        self.backend.execute('DROP TABLE dist_nexus')
        self.backend.execute('DROP TABLE IF EXISTS vars')
        self.backend.execute('DROP TABLE IF EXISTS source_lock')

    def _upgrade_to_22(self, oldvers):
        self._drop_triggers()
        self.backend.execute("ALTER TABLE file_tree ADD COLUMN maxcopies INTEGER")
        self.backend.execute("ALTER TABLE object ADD COLUMN maxcopies INTEGER")

    def _upgrade_to_23(self, oldvers):
        self._drop_triggers()
        self.backend.execute("ALTER TABLE nexus RENAME COLUMN count TO level")

    def _upgrade_to_24(self, oldvers):
        self.backend.execute("ALTER TABLE nexus ADD COLUMN disks TEXT NOT NULL DEFAULT ''")
        self.update_nexus_disk_text()

    def _upgrade_to_25(self, oldvers):
        self._drop_triggers()
        self.backend.execute('ALTER TABLE object RENAME TO old_object')
        self.backend.execute('ALTER TABLE file_tree RENAME TO old_file_tree')
        self.exec_schema(TABLES_INDICES)
        self.backend.execute('INSERT INTO object(hash, size, nexus, refs, priority, last_path, last_modtime) SELECT hash, size, nexus, refs, priority, last_path, last_modtime FROM old_object')
        self.backend.execute('INSERT INTO file_tree(virtual_path, last_modified, hash, ino, size, priority, metadata) SELECT virtual_path, last_modified, hash, ino, size, priority, metadata FROM old_file_tree')
        self.backend.execute('DROP TABLE old_object')
        self.backend.execute('DROP TABLE old_file_tree')

    def _upgrade_to_26(self, oldvers):
        self._drop_triggers()

    def _upgrade_to_27(self, oldvers):
        self._drop_triggers()
        # Add STRICT option to all tables
        self.alter_schema(
            ("type = 'table'", (), r'(\)\s*)(WITHOUT ROWID\s*)$', r'\1STRICT, \2'),
            ("type = 'table'", (), r'(\))(\s*)$', r'\1 STRICT\2'),
            debug=False
        )

    def _upgrade_to_28(self, oldvers):
        self._drop_triggers()

    def _upgrade_to_29(self, oldvers):
        self._drop_triggers()
        self.backend.execute("ALTER TABLE object ADD COLUMN saturated INTEGER GENERATED ALWAYS AS (CASE WHEN copies >= maxcopies THEN 1 ELSE 0 END)")
        self.backend.execute("ALTER TABLE object ADD COLUMN relevant_size INTEGER GENERATED ALWAYS AS (CASE WHEN refs > 0 THEN blocksize ELSE 0 END)")
        self.backend.execute("ALTER TABLE object ADD COLUMN saturated_size INTEGER GENERATED ALWAYS AS (CASE WHEN copies >= maxcopies AND refs > 0 THEN blocksize ELSE 0 END)")
        self.backend.execute("ALTER TABLE nexus ADD COLUMN saturated_size INTEGER NOT NULL DEFAULT 0")

    def _upgrade_to_30(self, oldvers):
        self._drop_triggers()
        self.backend.execute("UPDATE nexus SET saturated_size = COALESCE((SELECT SUM(saturated_size) FROM object o WHERE o.nexus = nexus.nexus), 0)")
        self.backend.execute("UPDATE nexus SET totalsize = COALESCE((SELECT SUM(relevant_size) FROM object o WHERE o.nexus = nexus.nexus), 0)")

    ##########################################################################################
    # Database
    ##########################################################################################

    def connect(self, *a, **kw):
        super().connect(*a, **kw)
        for n, f in SQL_FUNCS.items():
            self.backend.create_function(n, -1, f)

        self.disk_by_index = {}
        self.disk_by_name = {}
        self.disk_by_uuid = {}
        with self.execute('SELECT uuid, name, nexus_index, relative_path, size, fstype, fsuuid FROM disk WHERE name NOTNULL') as curs:
            for uuid, name, nexus_index, relative_path, size, fstype, fsuuid in curs:
                disk = BackupDisk(uuid, name, nexus_index, relative_path, size, fstype, fsuuid)
                self.disk_by_uuid[uuid] = disk
                self.disk_by_index[nexus_index] = disk
                self.disk_by_name[name] = disk

        self._disk_name_index = sorted((disk.name, index) for index, disk in self.disk_by_index.items())

        self.path_map = {
            virtual_path: Path(real_path)
            for virtual_path, real_path in
            self.query_list('SELECT virtual_path, real_path FROM path_map')
        }
        self.path_cache = self.path_map.copy()

    @in_transaction
    def copy_database(self, dest_dir=None):
        if dest_dir is None:
            dest_dir = self.current_disk.data_path

        self.execute_nonquery('PRAGMA wal_checkpoint(FULL)')
        if self.nocopydb:
            return

        if not self.current_disk or not self.current_disk.data_path:
            return

        progress = Progress.get_global()
        progress.set_status('copying database')
        tmppath = dest_dir / (DBNAME + '.new')
        backup_db = sqlite3.connect(tmppath)
        def log_progress(status, remaining, total):
            progress.progress((total - remaining) * 100 / total)

        self.backend.backup(backup_db, pages=256, progress=log_progress)
        backup_db.commit()
        backup_db.execute('PRAGMA wal_checkpoint(FULL)')
        backup_db.close()
        progress.reset()

        dstf = tmppath.with_name(DBNAME)
        tmppath.replace(dstf)

    def defer(self, f, *a, **kw):
        '''Queue a database-modifying function to run in the near future in a transaction.'''
        self.defer_queue.append((f, a, kw))

    @in_transaction
    def _run_deferred(self):
        for f, a, kw in self.defer_queue:
            f(*a, **kw)
        del self.defer_queue[:]

    def check_run_deferred(self, force=False):
        ctime = time.time()
        if force or ctime >= self.last_run_deferred + 0.2:
            self.last_run_deferred = ctime
            if self.defer_queue:
                self._run_deferred()

    def load_nexus(self):
        rtn = {}
        with self.execute('SELECT nexus, level, refs, totalsize, saturated_size, disks FROM nexus') as curs:
            for nexus, level, refs, totalsize, saturated_size, disks in curs:
                rtn[nexus] = Nexus(nexus, level, refs, totalsize, saturated_size, disks)
        return rtn

    def create_empty_nexus(self, nexusid:str):
        return Nexus(nexusid, nexus_level(nexusid), 0, 0, 0, ','.join(self.nexus_disks(nexusid)))

    def get_used_space(self, nexus_index):
        return self.query_scalar('SELECT COALESCE(sum(totalsize), 0) FROM nexus WHERE disk_in_nexus(nexus, ?)', (nexus_index,))

    def update_free_space(self):
        self.current_disk.used_space = self.get_used_space(self.current_disk.nexus_index)

    ##########################################################################################
    # Disk selection
    ##########################################################################################

    def set_backup_disk(self, arg):
        if '/' in arg:
            path = Path(arg)
            if not path.is_dir():
                raise DiskNotFoundError(f'path {path} does not exist or is not a directory')

            rdisk = BackupDisk.read_config(path)
            if not rdisk:
                raise DiskNotFoundError(f'path {path} does not contain a valid distbackup config')

            try:
                disk = self.disk_by_uuid[rdisk.uuid]
            except KeyError:
                raise ArgumentError(f'path {path} is from a different backup set') from None

            disk.data_path = path
            if disk.name != rdisk.name:
                disk.write_config()

            return self._set_current_disk(disk)

        try:
            return self.set_backup_disk_from_uuid(arg)
        except (ValueError, DiskNotFoundError):
            pass

        return self.set_backup_disk_from_name(arg)

    def set_backup_disk_from_name(self, name, search_data=True):
        try:
            disk = self.disk_by_name[name]
        except KeyError:
            raise DiskNotFoundError(f'disk {name} not found') from None

        self._set_current_disk(disk, search_data)

    def set_backup_disk_from_uuid(self, diskuuid, search_data=True):
        try:
            disk = self.disk_by_uuid[str(uuid.UUID(diskuuid))]
        except KeyError:
            raise DiskNotFoundError(f'disk {diskuuid} not found') from None

        self._set_current_disk(disk, search_data)

    def _set_current_disk(self, disk, search_data=True):
        self.current_disk = disk
        if not disk.data_path and search_data:
            BackupDisk.find_disk_data_paths([disk])

        log.debug(f'using data path {disk.data_path}')
        self.update_free_space()


    def require_disk(self):
        if not self.current_disk:
            raise ArgumentError('no disk selected')

    def require_disk_path(self):
        self.require_disk()
        if not self.current_disk.data_path:
            raise ArgumentError(f'{self.current_disk.name} is not mounted, or not configured correctly')

    def find_disk_data_paths(self):
        BackupDisk.find_disk_data_paths(self.disk_by_uuid.values())

    ##########################################################################################
    # Disk management
    ##########################################################################################

    @in_transaction('IMMEDIATE')
    def add_disk(self, name, path, size=None):
        uuid = new_uuid()
        new_disk = BackupDisk(uuid, name, None, size)
        new_disk.set_from_path(path)

        try:
            nexus_index = self.query_scalar('SELECT COALESCE((SELECT MIN(nexus_index) FROM disk WHERE name ISNULL), (SELECT MAX(nexus_index) + 1 FROM disk), 0)')
            new_disk.nexus_index = nexus_index
            self.execute_nonquery('DELETE FROM disk WHERE nexus_index = ?', (nexus_index,))
            self.execute_nonquery(
                'INSERT INTO disk(uuid, name, nexus_index, size, relative_path, fstype, fsuuid) VALUES(?, ?, ?, ?, ?, ?, ?)',
                (uuid, name, nexus_index, new_disk.size, new_disk.relative_path, new_disk.fstype, new_disk.fsuuid))

        except sqlite3.IntegrityError as e:
            raise ArgumentError(f'disk {name} already exists ({e})') from None

        self.current_disk = new_disk

        self.disk_by_index[nexus_index] = new_disk
        self.disk_by_name[name] = new_disk
        self.disk_by_uuid[uuid] = new_disk
        self._disk_name_index = sorted((disk.name, index) for index, disk in self.disk_by_index.items())

        self.update_nexus_disk_text()
        self.update_free_space()
        new_disk.write_config()
        new_disk.migrate_data()

        return new_disk


    @in_transaction
    def rename_disk(self, newname):
        self.require_disk()
        self.execute_nonquery('UPDATE disk SET name = ? WHERE uuid = ?', (newname, self.current_disk.uuid))
        self.disk_by_name.pop(self.current_disk.name, None)
        self.current_disk.name = newname
        self.disk_by_name[newname] = self.current_disk

        self.update_nexus_disk_text()

        if self.current_disk.data_path:
            self.current_disk.write_config()

    @in_transaction
    def set_disk_params(self, size=None, relpath=None, path=None, fstype=None, fsuuid=None):
        disk = self.current_disk
        if path is not None:
            disk.set_from_path(path)
            self.execute_nonquery(
                'UPDATE disk SET relative_path= ?, fstype = ?, fsuuid = ? WHERE uuid = ?',
                (disk.relative_path, disk.fstype, disk.fsuuid, disk.uuid)
            )

        if size is not None:
            self.execute_nonquery('UPDATE disk SET size = ? WHERE uuid = ?', (size, disk.uuid))
            disk.size = size

        if relpath is not None:
            disk.relative_path = relpath or None
            self.execute_nonquery('UPDATE disk SET relative_path = ? WHERE uuid = ?', (disk.relative_path, disk.uuid))

        if fstype is not None:
            disk.fstype = fstype or None
            self.execute_nonquery('UPDATE disk SET fstype = ? WHERE uuid = ?', (disk.fstype, disk.uuid))

        if fsuuid is not None:
            disk.fsuuid = fsuuid or None
            self.execute_nonquery('UPDATE disk SET fsuuid = ? WHERE uuid = ?', (disk.fsuuid, disk.uuid))

        self.update_nexus_disk_text()

        if disk.data_path:
            disk.write_config()

    def drop_disk_objects(self):
        index = self.current_disk.nexus_index
        with self.execute('SELECT nexus, level, totalsize FROM nexus') as curs:
            for nexus, copies, totalsize in curs:
                if disk_in_nexus(nexus, index):
                    new_nexus = nexus_without_disk(nexus, index)
                    self.execute_nonquery('UPDATE object SET nexus = ? WHERE nexus = ?', (new_nexus, nexus))

    @in_transaction
    def drop_disk(self):
        self.require_disk()
        self.drop_disk_objects()
        self.cleanrefs()

    @in_transaction
    def delete_disk(self):
        self.require_disk()

        used_space = self.query_scalar('SELECT sum(totalsize) FROM nexus WHERE disk_in_nexus(nexus, ?)', (self.current_disk.nexus_index,))
        if used_space:
            print(f'{self.current_disk.name} has backup files (use `disk drop`)')
            return

        self.execute_nonquery('UPDATE disk SET name = NULL WHERE uuid = ?', (self.current_disk.uuid,))
        self.cleanrefs()

    @in_transaction
    def _refresh_disk(self):
        self.require_disk()
        self.current_disk.migrate_data()
        self.drop_disk_objects()
        for hash, fpath, stat in self.current_disk.walk_data():
            log.info(f'found {hash}')
            self.execute_nonquery('INSERT OR IGNORE INTO object(hash, size) VALUES(?, ?)', (hash, stat.st_size))
            self.set_object_on_disk(hash, self.current_disk.nexus_index, True)

        self.cleanrefs()

    def refresh_disk(self):
        self._refresh_disk()
        self.update_free_space()
        self.copy_database()

    ##########################################################################################
    # Configuration
    ##########################################################################################

    @in_transaction
    def map_source(self, vpath:str, ppath:str):
        self.execute_nonquery('INSERT OR REPLACE INTO path_map(virtual_path, real_path) VALUES(?, ?)', (vpath, ppath))

    @in_transaction
    def unmap_source(self, vpath:str):
        rowcount, lastrowid = self.execute_nonquery('DELETE FROM path_map WHERE virtual_path = ?', (vpath,))
        return rowcount > 0

    @in_transaction
    def set_path_config(self, paths:list[str], updates:dict, clear:bool=False, append_keys=()):
        rtn = []
        for vpath in paths:
            vpath = canonical_path(vpath, True)
            prev_config_text = self.query_scalar('SELECT config FROM path_config WHERE virtual_path = ?', (vpath,))
            prev_config = json.loads(prev_config_text) if prev_config_text else {}

            if prev_config_text is None:
                config = updates.copy()
                if config:
                    self.execute_nonquery('INSERT INTO path_config(virtual_path, config) VALUES(?, ?)', (vpath, compact_json(config)))
            else:
                config = {} if clear else prev_config.copy()
                for key, val in updates.items():
                    if val is None or val == DEFAULT_CONFIG.get(key):
                        config.pop(key, None)
                    else:
                        if key in append_keys:
                            newval = list(config.get(key) or ())
                            for cval in val:
                                if cval not in newval:
                                    newval.append(cval)
                            val = newval
                        config[key] = val
                if config:
                    self.execute_nonquery('UPDATE path_config SET config = ? WHERE virtual_path = ?', (compact_json(config), vpath))
                else:
                    self.execute_nonquery("DELETE FROM path_config WHERE virtual_path = ?", (vpath,))

            rtn.append((vpath, prev_config, config))
        return rtn

    def export_config(self, path_prefix=None):
        path_config = {}
        disks = {}
        data = {
            'path_map': {
                k: str(v) for k, v in self.path_map.items()
            }
        }

        sql = 'SELECT virtual_path, config FROM path_config'
        args = ()
        if path_prefix:
            # Make sure the specified path is in the resulting export
            path_config[path_prefix] = {}
            text, args = prefix_sql('virtual_path', path_prefix)
            sql += " WHERE " + text

            data['path_prefix'] = path_prefix

        data['path_config'] = path_config
        for path, config in self.execute(sql, args):
            path_config[path] = json.loads(config)

        data['disks'] = disks
        with self.execute('SELECT uuid, name, relative_path, size, fstype, fsuuid FROM disk WHERE name NOTNULL') as curs:
            for uuid, name, relative_path, size, fstype, fsuuid in curs:
                disks[uuid] = {
                    'name': name,
                    'size_bytes': addcomma(size),
                    'relative_path': relative_path,
                    'fstype': fstype,
                    'fsuuid': fsuuid
                }

        return data

    @in_transaction
    def import_config(self, data):
        if 'path_map' in data:
            self.execute_nonquery('DELETE FROM path_map')
            for virtual_path, real_path in data['path_map'].items():
                if not isinstance(real_path, str):
                    raise ValueError(f'Expected string for path map: {virtual_path!r}, got {type(real_path).__name__!r}')

                virtual_path = canonical_path(virtual_path, True)
                self.execute_nonquery('INSERT INTO path_map(virtual_path, real_path) VALUES(?, ?)', (virtual_path, real_path))

        if 'path_config' in data:
            path_prefix = data.get('path_prefix')
            if path_prefix:
                text, args = prefix_sql('virtual_path', path_prefix)

                self.execute_nonquery("DELETE FROM path_config WHERE " + text, args)
            else:
                self.execute_nonquery('DELETE FROM path_config')

            for virtual_path, config in data['path_config'].items():
                virtual_path = canonical_path(virtual_path, True)
                if not isinstance(config, dict):
                    raise ValueError(f'Expected dict for path config: {virtual_path!r}, got {type(config).__name__}')

                for key, val in DEFAULT_CONFIG.items():
                    if key in config and type(val) != type(config[key]):
                        raise ValueError(f'Path {virtual_path}: Expected {type(val).__name__} for {key}, got {type(config[key]).__name__}')

                self.execute_nonquery('INSERT INTO path_config(virtual_path, config) VALUES(?, ?)', (virtual_path, compact_json(config)))

        if 'disks' in data:
            for uuid, config in data['disks'].items():
                realuuid = self.query_scalar('SELECT uuid FROM disk WHERE uuid = ?', (canonical_uuid(uuid),))
                if realuuid is None:
                    raise ValueError(f'Disk with UUID {realuuid} not in database (disks cannot be added with import)')

                if 'size_bytes' in config:
                    size = int(str(config['size_bytes']).replace(',', ''))
                    self.execute_nonquery('UPDATE disk SET size = ? WHERE uuid = ?', (size, realuuid))

                if 'relative_path' in config:
                    val = config['relative_path']
                    self.execute_nonquery('UPDATE disk SET relative_path = ? WHERE uuid = ?', (val, realuuid))

                if 'fstype' in config:
                    val = config['fstype']
                    self.execute_nonquery('UPDATE disk SET fstype = ? WHERE uuid = ?', (val, realuuid))

                if 'fsuuid' in config:
                    val = config['fsuuid']
                    self.execute_nonquery('UPDATE disk SET fsuuid = ? WHERE uuid = ?', (val, realuuid))

                if 'name' in config:
                    val = config['name']
                    self.execute_nonquery('UPDATE disk SET name = ? WHERE uuid = ?', (val, realuuid))

    ##########################################################################################
    # Utils
    ##########################################################################################

    def nexus_disks(self, nexus):
        '''Given a nexus ID, yields a sequence of the names of the disks in the nexus.'''
        for name, index in self._disk_name_index:
            if disk_in_nexus(nexus, index):
                yield name

    @in_transaction('IMMEDIATE')
    def cleanrefs(self):
        if not self.readonly:
            self.execute_nonquery('DELETE FROM nexus WHERE refs = 0 AND nexus != \'\'')
            self.execute_nonquery('DELETE FROM object WHERE refs = 0 AND copies = 0')

    @in_transaction
    def update_nexus_disk_text(self):
        updates = []
        disk_names = {}
        with self.execute('SELECT name, nexus_index FROM disk WHERE name NOTNULL') as curs:
            for name, index in curs:
                disk_names[index] = name

        with self.execute('SELECT nexus, disks FROM nexus') as curs:
            for nexus, text in curs:
                new_text = ','.join(sorted(name for i, c in enumerate(nexus) if (c == '1' and (name := disk_names.get(i)))))
                if new_text != text:
                    updates.append((new_text, nexus))
        if updates:
            with self.executemany('UPDATE nexus SET disks = ? WHERE nexus = ?', updates) as curs:
                pass

    def filter_files(self, pat=None, existing=(), sql='SELECT ft.virtual_path, ft.last_modified, ft.hash, ob.nexus FROM file_tree ft LEFT JOIN object ob ON ft.hash = ob.hash'):
        rx, prefix = pattern_to_regex(pat)

        args = ()
        if prefix:
            text, args = prefix_sql('ft.virtual_path', prefix)
            sql += r" WHERE " + text

        with self.execute(sql, args) as curs:
            for row in curs:
                path = row[0]
                if rx and not rx.match(path):
                    continue

                if any((base / path).exists() for base in existing):
                    continue

                yield row

    @in_transaction
    def set_object_on_disk(self, hash, diskid, present):
        prev_nexus = self.query_scalar('SELECT nexus FROM object WHERE hash = ?', (hash,))
        nexus = nexus_with_disk(prev_nexus, diskid) if present else nexus_without_disk(prev_nexus, diskid)
        if nexus != prev_nexus:
            self.execute_nonquery('UPDATE object SET nexus = ? WHERE hash = ?', (nexus, hash))
        return nexus

    @in_transaction
    def set_file_hash(self, vpath, hash):
        self.execute_nonquery('UPDATE file_tree SET hash = ? WHERE virtual_path = ?', (hash, vpath))

    ##########################################################################################
    # Path map
    ##########################################################################################

    def iter_paths(self):
        return self.path_map.items()

    def get_mapped_path(self, vpath:str):
        return self.path_map.get(vpath)

    def get_dir_path(self, dir):
        path = self.path_cache.get(dir + '/')
        if path is None:
            if not dir:
                return None
            parent, _, dirname = dir.rpartition('/')
            parent_path = self.get_dir_path(parent)
            if parent_path is None:
                return None
            path = self.path_cache[dir] = parent_path / dirname
        return path

    def get_file_path(self, fn):
        dir, _, filename = fn.rpartition('/')
        dir_path = self.get_dir_path(dir)
        if dir_path is None:
            return None
        return dir_path / filename



DBNAME = 'distbackup.sqlite'
META_SUFFIX = '.meta-inf'
DISK_CONF_NAME = 'distbackup-disk.json'
OBJECT_DIR_NAME = 'distbackup-objects'
LEGACY_OBJECT_DIR_NAME = '.distbackup.data'

DEFAULT_CONFIG = {
    'lock': False,
    'hide': False,
    'follow_symlinks': False,
    'include': [],
    'exclude': [],
}



SCHEMA_VERSION = 30

TABLES_INDICES = [
    # path_map: Maps virtual paths to actual directories on fhe filesystem
    """
    CREATE TABLE IF NOT EXISTS path_map (
        virtual_path TEXT PRIMARY KEY,
        real_path TEXT NOT NULL
    ) STRICT, WITHOUT ROWID
    """,

    # path_config: Contains configuration options for individual directories
    """
    CREATE TABLE IF NOT EXISTS path_config (
        virtual_path TEXT PRIMARY KEY,
        config TEXT NOT NULL DEFAULT '{}'
    ) STRICT, WITHOUT ROWID
    """,

    # file_tree: Contains information for each file that is backed up. Only leaf nodes
    # (i.e. files, not directories) have an entry in this table.
    """
    CREATE TABLE IF NOT EXISTS file_tree (
        -- Virtual path of the file, separated by "/"
        virtual_path TEXT PRIMARY KEY,

        -- Unix time in milliseconds
        last_modified INTEGER,

        -- SHA256 hash of the data. Empty if the hash is unknown, NULL if file is empty
        hash TEXT,

        -- Inode of the file (stat.st_ino). Used to quickly detect rename / move without having to rehash
        ino INTEGER,

        -- Size of the file, in bytes
        size INTEGER,

        -- Priority of the file. Files with a greater value for `priority` will be copied
        -- before lower priority files (assuming levels are equal)
        priority INTEGER NOT NULL DEFAULT 1,

        -- The maximum number of copies that this file can have on backup disks.
        maxcopies INTEGER,

        -- Compacted JSON of `.meta-inf` file associated with this file.
        metadata TEXT NOT NULL DEFAULT ''
    ) STRICT, WITHOUT ROWID
    """,

    "CREATE INDEX IF NOT EXISTS file_tree_hash ON file_tree(hash)",
    "CREATE INDEX IF NOT EXISTS file_need_hash ON file_tree(virtual_path) WHERE hash = ''",

    # symbolic_link: Contains all symbolic links from the backup set.
    """
    CREATE TABLE IF NOT EXISTS symbolic_link (
        virtual_path TEXT PRIMARY KEY,
        target TEXT NOT NULL
    ) STRICT, WITHOUT ROWID
    """,

    # nexus: A subset of the backup disks. Each row tracks the number and size of objects
    # stored on exactly those disks. Updated by triggers. The key is simply a binary
    # string of disks in little-endian order, stripped of trailing zeroes. '' represents
    # the empty set; '1' represents the nexus of only disk 0; '101' represents the nexus
    # of disk 0 with disk 2, etc.
    """
    CREATE TABLE IF NOT EXISTS nexus (
        -- Nexus key string, identifying which disks are in the set
        nexus TEXT PRIMARY KEY,

        -- Number of disks in the set
        level INTEGER GENERATED ALWAYS AS (length(replace(nexus, '0', ''))) STORED,

        -- Number of objects in this nexus. Updated by triggers.
        refs INTEGER DEFAULT 0,

        -- Sum of all `blocksize` columns of objects in this nexus. Updated by triggers.
        totalsize INTEGER NOT NULL DEFAULT 0,

        -- Names of disks in this nexus (informational, updated automatically)
        disks TEXT NOT NULL DEFAULT '',

        -- Sum of all `blocksize` columns of saturated objects in this nexus. Updated by triggers.
        saturated_size INTEGER NOT NULL DEFAULT 0
    ) STRICT, WITHOUT ROWID
    """,

    "CREATE INDEX IF NOT EXISTS nexus_prio ON nexus(level, totalsize DESC)",

    # object: Tracks which disks an object has been copied to. Multiple files with the
    # same content will reference the same object.
    """
    CREATE TABLE IF NOT EXISTS object (
        -- SHA256 hash of the object
        hash TEXT PRIMARY KEY,

        -- Exact size of the object in bytes
        size INTEGER DEFAULT 0,

        -- Size of the object, rounded up to next 4KB, plus 4KB for metadata
        blocksize INTEGER GENERATED ALWAYS AS (4096 + (size + 4095) & ~4095),

        -- Order of magnitude of the size
        size_order INTEGER GENERATED ALWAYS AS (CAST(floor(log2(blocksize + 0.01)) AS INTEGER)) STORED,

        -- Nexus of disks that this object is present on
        nexus TEXT NOT NULL DEFAULT '',

        -- Number of disks that this object exists on
        copies INTEGER GENERATED ALWAYS AS (length(replace(nexus, '0', ''))) STORED,

        -- Priority for this object. This is the maximum priority for any file_tree row
        -- that references this row.
        priority INTEGER NOT NULL DEFAULT 1,

        -- Maximum copies for this object. This is the maximum value for any file_tree row
        -- that references this row.
        maxcopies INTEGER,

        -- References to this object from the file_tree table, updated by triggers
        refs INTEGER NOT NULL DEFAULT 0,

        -- One of the paths that linked to this object. This is kept for
        -- informational purposes, so that if a file's content changes and causes
        -- this object to get deleted, we can show the user which file it was, along
        -- with the modification date.
        last_path TEXT,
        last_modtime INTEGER,

        -- This will be 1 iff the object is at the maximum number of copies.
        saturated INTEGER GENERATED ALWAYS AS (CASE WHEN copies >= maxcopies THEN 1 ELSE 0 END),

        -- This is equal to the blocksize iff the object is referred to by the file_tree table.
        relevant_size INTEGER GENERATED ALWAYS AS (CASE WHEN refs > 0 THEN blocksize ELSE 0 END),

        -- This is equal to the blocksize iff the object is referred to by the file_tree table.
        saturated_size INTEGER GENERATED ALWAYS AS (CASE WHEN copies >= maxcopies AND refs > 0 THEN blocksize ELSE 0 END)

    ) STRICT
    """,

    "CREATE INDEX IF NOT EXISTS object_prio ON object(copies DESC, blocksize)",
    "CREATE INDEX IF NOT EXISTS object_size ON object(blocksize)",
    "CREATE INDEX IF NOT EXISTS object_refs ON object(refs)",
    "CREATE INDEX IF NOT EXISTS object_refs_copies ON object(refs, copies)",
    "CREATE INDEX IF NOT EXISTS object_nexus ON object(nexus)",

    # disk: Represents a backup disk that stores objects.
    """
    CREATE TABLE IF NOT EXISTS disk (
        -- UUID identifying the disk. This is generated when the disk is added and
        -- never changes.
        uuid TEXT PRIMARY KEY,

        -- Name of this disk. NULL if the disk has been deleted. Deleted disks are
        -- kept so the `nexus_index` can be reused.
        name TEXT UNIQUE,

        -- Index of this disk within nexus indetifiers.
        nexus_index INTEGER NOT NULL UNIQUE,

        -- Path to object files on the disk realtive to its mountpoint.
        relative_path TEXT,

        -- If not NULL, only search on partitions matching the given fstype
        fstype TEXT,

        -- If not NULL, only search on partitions matching the given UUID
        fsuuid TEXT,

        -- Size of the disk in bytes.
        size INTEGER

    ) STRICT, WITHOUT ROWID
    """,

    "CREATE INDEX IF NOT EXISTS unused_nexus_index ON disk(nexus_index) WHERE name ISNULL",
]

TRIGGERS = [
    # Nexus update triggers. min(NEW.refs, 1) is used so that objects that aren't
    # referred to in the file_tree don't count as being present.
    """
    CREATE TRIGGER IF NOT EXISTS object_insert
    AFTER INSERT ON object
    BEGIN
        INSERT INTO nexus(nexus, refs, totalsize, saturated_size)
          VALUES(NEW.nexus, 1, NEW.relevant_size, NEW.saturated_size)
        ON CONFLICT(nexus) DO UPDATE SET
          refs = refs + 1,
          totalsize = totalsize + NEW.relevant_size,
          saturated_size = saturated_size + NEW.saturated_size;
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS object_update
    AFTER UPDATE ON object
    WHEN OLD.relevant_size != NEW.relevant_size OR OLD.saturated_size != NEW.saturated_size OR OLD.nexus != NEW.nexus
    BEGIN
        INSERT INTO nexus(nexus, refs, totalsize, saturated_size)
          VALUES(NEW.nexus, 1, NEW.relevant_size, NEW.saturated_size)
        ON CONFLICT(nexus) DO UPDATE SET
          refs = refs + 1,
          totalsize = totalsize + NEW.relevant_size,
          saturated_size = saturated_size + NEW.saturated_size;

        UPDATE nexus SET
          refs = refs - 1,
          totalsize = totalsize - OLD.relevant_size,
          saturated_size = saturated_size - OLD.saturated_size
        WHERE nexus = OLD.nexus;
    END
    """,
    """
    CREATE TRIGGER IF NOT EXISTS object_delete
    BEFORE DELETE ON object
    BEGIN
        UPDATE nexus SET
          refs = refs - 1,
          totalsize = totalsize - OLD.relevant_size,
          saturated_size = saturated_size - OLD.saturated_size
        WHERE nexus = OLD.nexus;
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS file_tree_insert
    AFTER INSERT ON file_tree
    WHEN NEW.hash <> ''
    BEGIN
        INSERT INTO object(hash, size, priority, maxcopies, refs, last_path, last_modtime)
          VALUES(NEW.hash, NEW.size, NEW.priority, NEW.maxcopies, 1, NEW.virtual_path, NEW.last_modified)
        ON CONFLICT(hash) DO UPDATE SET
          refs = refs + 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = NEW.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = NEW.hash);
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS file_tree_update
    AFTER UPDATE ON file_tree
    WHEN OLD.hash != NEW.hash
    BEGIN
        INSERT INTO object(hash, size, priority, maxcopies, refs, last_path, last_modtime)
          VALUES(NEW.hash, NEW.size, NEW.priority, NEW.maxcopies, 1, NEW.virtual_path, NEW.last_modified)
        ON CONFLICT(hash) DO UPDATE SET
          refs = refs + 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = NEW.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = NEW.hash);

        UPDATE object SET
          refs = refs - 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = OLD.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = OLD.hash)
        WHERE hash = OLD.hash;
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS file_tree_delete
    BEFORE DELETE ON file_tree
    BEGIN
        UPDATE object SET
          refs = refs - 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = OLD.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = OLD.hash)
        WHERE hash = OLD.hash;
    END
    """
]

PRE_DATA = [
    # Make sure there is always a nexus entry representing the empty set
    "INSERT OR IGNORE INTO nexus(nexus) VALUES ('')",
]



import os
import logging

from typing import Optional
from collections import deque
from pathlib import Path

from .utils import check_dir, restore_mtime
from .database import BackupDB
from .sourcefile import SourceFile
log = logging.getLogger(__name__)

class Restore:
    def __init__(self, db:BackupDB, restore_path:Optional[Path]=None, sync:bool=False):
        self.db:BackupDB = db
        self.sync:bool = sync
        self.restore_path:Optional[Path] = restore_path

    def _restore_one(self, f, hashpath, hash, path, lastmod):
        log.info(f'{hash} -> {path}')
        try:
            dpath = self.restore_path / path
            check_dir(dpath)
            f(hashpath, dpath)
        except OSError as e:
            log.error(f'Error restoring {path}: {e}')

        try:
            tm = restore_mtime(lastmod)
            os.utime(dpath, (tm, tm))
        except OSError as e:
            pass

    def _restore(self, f, pat=None, existing=()):
        self.db.require_disk_path()
        self.db.current_disk.migrate_data()
        not_present = []
        existing = [Path(v) for v in existing]
        if self.restore_path:
            existing.append(self.restore_path)

        for path, lastmod, hash, nexus in self.db.filter_files(pat, existing):
            path = path.lstrip('/')
            if hash is None:
                dpath = self.restore_path / path
                check_dir(dpath)
                dpath.write_bytes(b'')
                continue

            hashpath = self.db.current_disk.hashpath(hash)
            if hashpath.exists():
                self._restore_one(f, hashpath, hash, path, lastmod)
            elif nexus:
                not_present.append(path)


        for path, dest in self.db.filter_files(pat, existing, 'SELECT ft.virtual_path, ft.target FROM symbolic_link AS ft'):
            try:
                dpath = self.restore_path / path
                check_dir(dpath)
                dpath.symlink_to(dest)

            except OSError as e:
                log.error(f'Error restoring link {path}: {e}')

        if not_present:
            log.info('')
            log.info('The following files still need to be restored:')
            for file in not_present:
                log.info(f'  {file}')
            log.info(f'total not present: {len(not_present)}')

    def restore_link(self, pat=None, existing=()):
        self._restore(os.link, pat, existing)

    def restore_copy(self, pat=None, existing=()):
        def f(fpath, dpath):
            srcf = SourceFile(fpath)
            tmppath = dpath.with_name(f'{dpath.name}.dbtmp~')
            try:
                srcf.setdest(tmppath)
                srcf.copy()
                tmppath.replace(dpath)
            finally:
                if tmppath.exists():
                    tmppath.unlink()

        self._restore(f, pat, existing)

    def restore_set(self, pat=None, existing=()):
        matching_nexus = set()
        existing = [Path(v) for v in existing]
        disk_names = {i: disk.name for i, disk in self.db.disk_by_index.items()}

        sql = 'SELECT ft.virtual_path, ft.last_modified, ft.hash, f.nexus FROM file_tree AS ft INNER JOIN object AS f ON ft.hash = f.hash'
        filecount = 0
        for path, lastmod, hash, nexus in self.db.filter_files(pat, existing, sql):
            if nexus:
                filecount += 1
                matching_nexus.add(nexus)

        if not filecount:
            log.info(f'No files match "{pat}".')
            return

        max_disk = max(len(z) for z in matching_nexus)
        disk_seq = range(0, max_disk)

        def build_rtn(disks, cnt, sets):
            for j, next_set in disk_set:
                log.info(f'{j+1:5d} {len(next_set):5d} {next_set}')

        log.info(f'Finding minimum set for {filecount} files matching "{pat}" ...')

        all_file_nexus = set(sum((1<<i) for i, c in enumerate(nexus) if c == '1') for nexus in matching_nexus)
        for z in sorted(all_file_nexus):
            log.info(' '.join((disk_names.get(j, '?') if (z&(1<<j)) else '--').ljust(10) for j in disk_seq))

        found_list = None
        search_queue = deque()
        search_queue.append((0, all_file_nexus))
        while search_queue:
            chosen_disks, sets = search_queue.popleft()
            disk_set = [(disk_id, set()) for disk_id in disk_seq if not chosen_disks & (1<<disk_id)]
            for disk_id, next_set in disk_set:
                mask = 1 << disk_id
                if mask & chosen_disks:
                    continue
                for nexus_set in sets:
                    if not nexus_set & mask:
                        next_set.add(nexus_set & ~mask)

            disk_set.sort(key=lambda v: len(v[1]))
            del disk_set[4:]
            for disk_id, next_set in disk_set:
                new_chosen = chosen_disks | (1 << disk_id)
                if not next_set:
                    found_list = new_chosen
                    break
                search_queue.append((new_chosen, next_set))
            if found_list is not None:
                break

        log.info('')
        log.info('The smallest set of disks that contains the files is:')
        log.info('    ' + ' '.join([disk_names.get(n, '?') for n in disk_seq if found_list & (1 << n)]))
